@@ -19,7 +19,6 @@
 using mozilla::AssertedCast;
 using mozilla::GenericPromise;
 using mozilla::LogLevel;
-using mozilla::PRemoteSpellcheckEngineChild;
 using mozilla::RemoteSpellcheckEngineChild;
 using mozilla::TextServicesDocument;
 using mozilla::dom::ContentChild;
@@ -31,9 +30,6 @@ static mozilla::LazyLogModule sSpellChecker("SpellChecker");
 NS_IMPL_CYCLE_COLLECTION(mozSpellChecker, mTextServicesDocument,
                          mPersonalDictionary)
 
-NS_IMPL_CYCLE_COLLECTION_ROOT_NATIVE(mozSpellChecker, AddRef)
-NS_IMPL_CYCLE_COLLECTION_UNROOT_NATIVE(mozSpellChecker, Release)
-
 mozSpellChecker::mozSpellChecker() : mEngine(nullptr) {}
 
 mozSpellChecker::~mozSpellChecker() {
@@ -144,18 +140,14 @@ RefPtr<mozilla::CheckWordPromise> mozSpellChecker::CheckWords(
 
 nsresult mozSpellChecker::CheckWord(const nsAString& aWord, bool* aIsMisspelled,
                                     nsTArray<nsString>* aSuggestions) {
+  if (XRE_IsContentProcess()) {
+    // Use async version (CheckWords or Suggest) on content process
+    return NS_ERROR_FAILURE;
+  }
+
   nsresult result;
   bool correct;
 
-  if (XRE_IsContentProcess()) {
-    MOZ_ASSERT(aSuggestions, "Use CheckWords if content process");
-    if (!mEngine->SendCheckAndSuggest(nsString(aWord), aIsMisspelled,
-                                      aSuggestions)) {
-      return NS_ERROR_NOT_AVAILABLE;
-    }
-    return NS_OK;
-  }
-
   if (!mSpellCheckingEngine) {
     return NS_ERROR_NULL_POINTER;
   }
@@ -172,6 +164,46 @@ nsresult mozSpellChecker::CheckWord(const nsAString& aWord, bool* aIsMisspelled,
   return NS_OK;
 }
 
+RefPtr<mozilla::SuggestionsPromise> mozSpellChecker::Suggest(
+    const nsAString& aWord, uint32_t aMaxCount) {
+  if (XRE_IsContentProcess()) {
+    return mEngine->SendSuggest(aWord, aMaxCount)
+        ->Then(
+            mozilla::GetCurrentSerialEventTarget(), __func__,
+            [](nsTArray<nsString>&& aSuggestions) {
+              return mozilla::SuggestionsPromise::CreateAndResolve(
+                  std::move(aSuggestions), __func__);
+            },
+            [](mozilla::ipc::ResponseRejectReason&& aReason) {
+              return mozilla::SuggestionsPromise::CreateAndReject(
+                  NS_ERROR_NOT_AVAILABLE, __func__);
+            });
+  }
+
+  if (!mSpellCheckingEngine) {
+    return mozilla::SuggestionsPromise::CreateAndReject(NS_ERROR_NOT_AVAILABLE,
+                                                        __func__);
+  }
+
+  bool correct;
+  nsresult rv = mSpellCheckingEngine->Check(aWord, &correct);
+  if (NS_FAILED(rv)) {
+    return mozilla::SuggestionsPromise::CreateAndReject(rv, __func__);
+  }
+  nsTArray<nsString> suggestions;
+  if (!correct) {
+    rv = mSpellCheckingEngine->Suggest(aWord, suggestions);
+    if (NS_FAILED(rv)) {
+      return mozilla::SuggestionsPromise::CreateAndReject(rv, __func__);
+    }
+    if (suggestions.Length() > aMaxCount) {
+      suggestions.TruncateLength(aMaxCount);
+    }
+  }
+  return mozilla::SuggestionsPromise::CreateAndResolve(std::move(suggestions),
+                                                       __func__);
+}
+
 nsresult mozSpellChecker::Replace(const nsAString& aOldWord,
                                   const nsAString& aNewWord,
                                   bool aAllOccurrences) {
@@ -207,6 +239,9 @@ nsresult mozSpellChecker::Replace(const nsAString& aOldWord,
   }
   int32_t currOffset = 0;
   int32_t currentBlock = 0;
+  int32_t wordLengthDifference =
+      AssertedCast<int32_t>(static_cast<int64_t>(aNewWord.Length()) -
+                            static_cast<int64_t>(aOldWord.Length()));
   while (NS_SUCCEEDED(mTextServicesDocument->IsDone(&done)) && !done) {
     nsAutoString str;
     mTextServicesDocument->GetCurrentTextBlock(str);
@@ -215,18 +250,25 @@ nsresult mozSpellChecker::Replace(const nsAString& aOldWord,
         // if we are before the current selection point but in the same
         // block move the selection point forwards
         if (currentBlock == startBlock && begin < selOffset) {
-          selOffset += int32_t(aNewWord.Length()) - int32_t(aOldWord.Length());
+          selOffset += wordLengthDifference;
           if (selOffset < begin) {
             selOffset = begin;
           }
         }
-        MOZ_KnownLive(mTextServicesDocument)
-            ->SetSelection(AssertedCast<uint32_t>(begin),
-                           AssertedCast<uint32_t>(end - begin));
-        MOZ_KnownLive(mTextServicesDocument)->InsertText(aNewWord);
+        // Don't keep running if selecting or inserting text fails because
+        // it may cause infinite loop.
+        if (NS_WARN_IF(NS_FAILED(
+                MOZ_KnownLive(mTextServicesDocument)
+                    ->SetSelection(AssertedCast<uint32_t>(begin),
+                                   AssertedCast<uint32_t>(end - begin))))) {
+          return NS_ERROR_FAILURE;
+        }
+        if (NS_WARN_IF(NS_FAILED(
+                MOZ_KnownLive(mTextServicesDocument)->InsertText(aNewWord)))) {
+          return NS_ERROR_FAILURE;
+        }
         mTextServicesDocument->GetCurrentTextBlock(str);
-        end += (aNewWord.Length() -
-                aOldWord.Length());  // recursion was cute in GEB, not here.
+        end += wordLengthDifference;  // recursion was cute in GEB, not here.
       }
       currOffset = end;
     }
@@ -350,31 +392,31 @@ nsresult mozSpellChecker::GetDictionaryList(
   return NS_OK;
 }
 
-nsresult mozSpellChecker::GetCurrentDictionary(nsACString& aDictionary) {
+nsresult mozSpellChecker::GetCurrentDictionaries(
+    nsTArray<nsCString>& aDictionaries) {
   if (XRE_IsContentProcess()) {
-    aDictionary = mCurrentDictionary;
+    aDictionaries = mCurrentDictionaries.Clone();
     return NS_OK;
   }
 
   if (!mSpellCheckingEngine) {
-    aDictionary.Truncate();
+    aDictionaries.Clear();
     return NS_OK;
   }
 
-  return mSpellCheckingEngine->GetDictionary(aDictionary);
+  return mSpellCheckingEngine->GetDictionaries(aDictionaries);
 }
 
 nsresult mozSpellChecker::SetCurrentDictionary(const nsACString& aDictionary) {
   if (XRE_IsContentProcess()) {
-    nsCString wrappedDict = nsCString(aDictionary);
+    mCurrentDictionaries.Clear();
     bool isSuccess;
-    mEngine->SendSetDictionary(wrappedDict, &isSuccess);
+    mEngine->SendSetDictionary(aDictionary, &isSuccess);
     if (!isSuccess) {
-      mCurrentDictionary.Truncate();
       return NS_ERROR_NOT_AVAILABLE;
     }
 
-    mCurrentDictionary = wrappedDict;
+    mCurrentDictionaries.AppendElement(aDictionary);
     return NS_OK;
   }
 
@@ -392,18 +434,19 @@ nsresult mozSpellChecker::SetCurrentDictionary(const nsACString& aDictionary) {
   rv = GetEngineList(&spellCheckingEngines);
   NS_ENSURE_SUCCESS(rv, rv);
 
+  nsTArray<nsCString> dictionaries;
+  dictionaries.AppendElement(aDictionary);
   for (int32_t i = 0; i < spellCheckingEngines.Count(); i++) {
-    // We must set mSpellCheckingEngine before we call SetDictionary, since
-    // SetDictionary calls back to this spell checker to check if the
+    // We must set mSpellCheckingEngine before we call SetDictionaries, since
+    // SetDictionaries calls back to this spell checker to check if the
     // dictionary was set
     mSpellCheckingEngine = spellCheckingEngines[i];
-
-    rv = mSpellCheckingEngine->SetDictionary(aDictionary);
+    rv = mSpellCheckingEngine->SetDictionaries(dictionaries);
 
     if (NS_SUCCEEDED(rv)) {
       nsCOMPtr<mozIPersonalDictionary> personalDictionary =
           do_GetService("@mozilla.org/spellchecker/personaldictionary;1");
-      mSpellCheckingEngine->SetPersonalDictionary(personalDictionary.get());
+      mSpellCheckingEngine->SetPersonalDictionary(personalDictionary);
 
       mConverter = new mozEnglishWordUtils;
       return NS_OK;
@@ -416,6 +459,59 @@ nsresult mozSpellChecker::SetCurrentDictionary(const nsACString& aDictionary) {
   return NS_ERROR_NOT_AVAILABLE;
 }
 
+RefPtr<GenericPromise> mozSpellChecker::SetCurrentDictionaries(
+    const nsTArray<nsCString>& aDictionaries) {
+  if (XRE_IsContentProcess()) {
+    if (!mEngine) {
+      mCurrentDictionaries.Clear();
+      return GenericPromise::CreateAndReject(NS_ERROR_NOT_AVAILABLE, __func__);
+    }
+
+    // mCurrentDictionaries will be set by RemoteSpellCheckEngineChild
+    return mEngine->SetCurrentDictionaries(aDictionaries);
+  }
+
+  // Calls to mozISpellCheckingEngine::SetDictionary might destroy us
+  RefPtr<mozSpellChecker> kungFuDeathGrip = this;
+
+  mSpellCheckingEngine = nullptr;
+
+  if (aDictionaries.IsEmpty()) {
+    return GenericPromise::CreateAndResolve(true, __func__);
+  }
+
+  nsresult rv;
+  nsCOMArray<mozISpellCheckingEngine> spellCheckingEngines;
+  rv = GetEngineList(&spellCheckingEngines);
+  if (NS_FAILED(rv)) {
+    return GenericPromise::CreateAndReject(rv, __func__);
+  }
+
+  for (int32_t i = 0; i < spellCheckingEngines.Count(); i++) {
+    // We must set mSpellCheckingEngine before we call SetDictionaries, since
+    // SetDictionaries calls back to this spell checker to check if the
+    // dictionary was set
+    mSpellCheckingEngine = spellCheckingEngines[i];
+    rv = mSpellCheckingEngine->SetDictionaries(aDictionaries);
+
+    if (NS_SUCCEEDED(rv)) {
+      mCurrentDictionaries = aDictionaries.Clone();
+
+      nsCOMPtr<mozIPersonalDictionary> personalDictionary =
+          do_GetService("@mozilla.org/spellchecker/personaldictionary;1");
+      mSpellCheckingEngine->SetPersonalDictionary(personalDictionary);
+
+      mConverter = new mozEnglishWordUtils;
+      return GenericPromise::CreateAndResolve(true, __func__);
+    }
+  }
+
+  mSpellCheckingEngine = nullptr;
+
+  // We could not find any engine with the requested dictionary
+  return GenericPromise::CreateAndReject(NS_ERROR_NOT_AVAILABLE, __func__);
+}
+
 RefPtr<GenericPromise> mozSpellChecker::SetCurrentDictionaryFromList(
     const nsTArray<nsCString>& aList) {
   if (aList.IsEmpty()) {
@@ -423,7 +519,12 @@ RefPtr<GenericPromise> mozSpellChecker::SetCurrentDictionaryFromList(
   }
 
   if (XRE_IsContentProcess()) {
-    // mCurrentDictionary will be set by RemoteSpellCheckEngineChild
+    if (!mEngine) {
+      mCurrentDictionaries.Clear();
+      return GenericPromise::CreateAndReject(NS_ERROR_NOT_AVAILABLE, __func__);
+    }
+
+    // mCurrentDictionaries will be set by RemoteSpellCheckEngineChild
     return mEngine->SetCurrentDictionaryFromList(aList);
   }
 
@@ -456,7 +557,12 @@ nsresult mozSpellChecker::SetupDoc(int32_t* outBlockOffset) {
         // S begins or ends in TB but extends outside of TB.
         case TextServicesDocument::BlockSelectionStatus::eBlockPartial:
           // the TS doc points to the block we want.
-          MOZ_ASSERT(selOffset != UINT32_MAX || selLength != UINT32_MAX);
+          if (NS_WARN_IF(selOffset == UINT32_MAX) ||
+              NS_WARN_IF(selLength == UINT32_MAX)) {
+            rv = mTextServicesDocument->FirstBlock();
+            *outBlockOffset = 0;
+            break;
+          }
           *outBlockOffset = AssertedCast<int32_t>(selOffset + selLength);
           break;
 
@@ -469,7 +575,12 @@ nsresult mozSpellChecker::SetupDoc(int32_t* outBlockOffset) {
 
         // TB contains entire S.
         case TextServicesDocument::BlockSelectionStatus::eBlockContains:
-          MOZ_ASSERT(selOffset != UINT32_MAX || selLength != UINT32_MAX);
+          if (NS_WARN_IF(selOffset == UINT32_MAX) ||
+              NS_WARN_IF(selLength == UINT32_MAX)) {
+            rv = mTextServicesDocument->FirstBlock();
+            *outBlockOffset = 0;
+            break;
+          }
           *outBlockOffset = AssertedCast<int32_t>(selOffset + selLength);
           break;
 
