@@ -6,15 +6,15 @@
 
 #include "GLReadTexImageHelper.h"
 
-#include "gfx2DGlue.h"
-#include "gfxColor.h"
-#include "gfxTypes.h"
+#include <utility>
+
 #include "GLContext.h"
 #include "OGLShaderProgram.h"
 #include "ScopedGLHelpers.h"
-
+#include "gfx2DGlue.h"
+#include "gfxColor.h"
+#include "gfxTypes.h"
 #include "mozilla/gfx/2D.h"
-#include "mozilla/Move.h"
 
 namespace mozilla {
 namespace gl {
@@ -220,7 +220,7 @@ void SwapRAndBComponents(DataSourceSurface* surf) {
     const uint8_t* rowEnd = row + rowBytes;
 
     while (row != rowEnd) {
-      Swap(row[0], row[2]);
+      std::swap(row[0], row[2]);
       row += 4;
     }
 
@@ -382,8 +382,6 @@ void ReadPixelsIntoDataSurface(GLContext* gl, DataSourceSurface* dest) {
   map = Nothing();
 
   if (readSurf != dest) {
-    MOZ_ASSERT(readFormat == LOCAL_GL_RGBA);
-    MOZ_ASSERT(readType == LOCAL_GL_UNSIGNED_BYTE);
     gfx::Factory::CopyDataSourceSurface(readSurf, dest);
   }
 }
@@ -424,7 +422,6 @@ already_AddRefed<DataSourceSurface> ReadBackSurface(GLContext* gl,
                                                     bool aYInvert,
                                                     SurfaceFormat aFormat) {
   gl->MakeCurrent();
-  gl->GuaranteeResolve();
   gl->fActiveTexture(LOCAL_GL_TEXTURE0);
   gl->fBindTexture(LOCAL_GL_TEXTURE_2D, aTexture);
 
