@@ -7,21 +7,23 @@
 #include "OGLShaderProgram.h"
 
 #include <stdint.h>  // for uint32_t
-#include <sstream>   // for ostringstream
+
+#include <sstream>  // for std::ostringstream
+
+#include "GLContext.h"
 #include "gfxEnv.h"
 #include "gfxRect.h"  // for gfxRect
 #include "gfxUtils.h"
-#include "mozilla/DebugOnly.h"          // for DebugOnly
+#include "mozilla/DebugOnly.h"  // for DebugOnly
+#include "mozilla/gfx/Logging.h"
 #include "mozilla/layers/Compositor.h"  // for BlendOpIsMixBlendMode
 #include "nsAString.h"
 #include "nsString.h"  // for nsAutoCString
-#include "Layers.h"
-#include "GLContext.h"
 
 namespace mozilla {
 namespace layers {
 
-using namespace std;
+using std::endl;
 
 #define GAUSSIAN_KERNEL_HALF_WIDTH 11
 #define GAUSSIAN_KERNEL_STEP 0.2
@@ -42,25 +44,14 @@ static void AddUniforms(ProgramProfileOGL& aProfile) {
                                              "uYTexture",
                                              "uCbTexture",
                                              "uCrTexture",
-                                             "uBlackTexture",
-                                             "uWhiteTexture",
-                                             "uMaskTexture",
-                                             "uBackdropTexture",
                                              "uRenderColor",
                                              "uTexCoordMultiplier",
                                              "uCbCrTexCoordMultiplier",
-                                             "uMaskCoordMultiplier",
-                                             "uTexturePass2",
-                                             "uColorMatrix",
-                                             "uColorMatrixVector",
-                                             "uBlurRadius",
-                                             "uBlurOffset",
-                                             "uBlurAlpha",
-                                             "uBlurGaussianKernel",
                                              "uSSEdges",
                                              "uViewportSize",
                                              "uVisibleCenter",
                                              "uYuvColorMatrix",
+                                             "uYuvOffsetVector",
                                              nullptr};
 
   for (int i = 0; sKnownUniformNames[i] != nullptr; ++i) {
@@ -120,6 +111,9 @@ void ShaderConfigOGL::SetColorMultiplier(uint32_t aMultiplier) {
 void ShaderConfigOGL::SetNV12(bool aEnabled) {
   SetFeature(ENABLE_TEXTURE_NV12, aEnabled);
   MOZ_ASSERT(!(mFeatures & ENABLE_TEXTURE_YCBCR));
+#ifdef MOZ_WAYLAND
+  SetFeature(ENABLE_TEXTURE_NV12_GA_SWITCH, aEnabled);
+#endif
 }
 
 void ShaderConfigOGL::SetComponentAlpha(bool aEnabled) {
@@ -157,7 +151,7 @@ void ShaderConfigOGL::SetDynamicGeometry(bool aEnabled) {
 /* static */
 ProgramProfileOGL ProgramProfileOGL::GetProfileFor(ShaderConfigOGL aConfig) {
   ProgramProfileOGL result;
-  ostringstream fs, vs;
+  std::ostringstream fs, vs;
 
   AddUniforms(result);
 
@@ -185,7 +179,7 @@ ProgramProfileOGL ProgramProfileOGL::GetProfileFor(ShaderConfigOGL aConfig) {
     vs << "attribute vec2 aCoord;" << endl;
   }
 
-  result.mAttributes.AppendElement(Pair<nsCString, GLuint>{"aCoord", 0});
+  result.mAttributes.AppendElement(std::pair<nsCString, GLuint>{"aCoord", 0});
 
   if (!(aConfig.mFeatures & ENABLE_RENDER_COLOR)) {
     vs << "uniform mat4 uTextureTransform;" << endl;
@@ -194,7 +188,8 @@ ProgramProfileOGL ProgramProfileOGL::GetProfileFor(ShaderConfigOGL aConfig) {
 
     if (aConfig.mFeatures & ENABLE_DYNAMIC_GEOMETRY) {
       vs << "attribute vec2 aTexCoord;" << endl;
-      result.mAttributes.AppendElement(Pair<nsCString, GLuint>{"aTexCoord", 1});
+      result.mAttributes.AppendElement(
+          std::pair<nsCString, GLuint>{"aTexCoord", 1});
     }
   }
 
@@ -399,10 +394,12 @@ ProgramProfileOGL ProgramProfileOGL::GetProfileFor(ShaderConfigOGL aConfig) {
     fs << "uniform " << sampler2D << " uCbTexture;" << endl;
     fs << "uniform " << sampler2D << " uCrTexture;" << endl;
     fs << "uniform mat3 uYuvColorMatrix;" << endl;
+    fs << "uniform vec3 uYuvOffsetVector;" << endl;
   } else if (aConfig.mFeatures & ENABLE_TEXTURE_NV12) {
     fs << "uniform " << sampler2D << " uYTexture;" << endl;
     fs << "uniform " << sampler2D << " uCbTexture;" << endl;
     fs << "uniform mat3 uYuvColorMatrix;" << endl;
+    fs << "uniform vec3 uYuvOffsetVector;" << endl;
   } else if (aConfig.mFeatures & ENABLE_TEXTURE_COMPONENT_ALPHA) {
     fs << "uniform " << sampler2D << " uBlackTexture;" << endl;
     fs << "uniform " << sampler2D << " uWhiteTexture;" << endl;
@@ -458,23 +455,32 @@ ProgramProfileOGL ProgramProfileOGL::GetProfileFor(ShaderConfigOGL aConfig) {
              << "(uYTexture, coord * uTexCoordMultiplier).r;" << endl;
           fs << "  COLOR_PRECISION float cb = " << texture2D
              << "(uCbTexture, coord * uCbCrTexCoordMultiplier).r;" << endl;
-          fs << "  COLOR_PRECISION float cr = " << texture2D
-             << "(uCbTexture, coord * uCbCrTexCoordMultiplier).a;" << endl;
+          if (aConfig.mFeatures & ENABLE_TEXTURE_NV12_GA_SWITCH) {
+            fs << "  COLOR_PRECISION float cr = " << texture2D
+               << "(uCbTexture, coord * uCbCrTexCoordMultiplier).g;" << endl;
+          } else {
+            fs << "  COLOR_PRECISION float cr = " << texture2D
+               << "(uCbTexture, coord * uCbCrTexCoordMultiplier).a;" << endl;
+          }
         } else {
           fs << "  COLOR_PRECISION float y = " << texture2D
              << "(uYTexture, coord).r;" << endl;
           fs << "  COLOR_PRECISION float cb = " << texture2D
              << "(uCbTexture, coord).r;" << endl;
-          fs << "  COLOR_PRECISION float cr = " << texture2D
-             << "(uCbTexture, coord).a;" << endl;
+          if (aConfig.mFeatures & ENABLE_TEXTURE_NV12_GA_SWITCH) {
+            fs << "  COLOR_PRECISION float cr = " << texture2D
+               << "(uCbTexture, coord).g;" << endl;
+          } else {
+            fs << "  COLOR_PRECISION float cr = " << texture2D
+               << "(uCbTexture, coord).a;" << endl;
+          }
         }
       }
       fs << "  vec3 yuv = vec3(y, cb, cr);" << endl;
       if (aConfig.mMultiplier != 1) {
         fs << "  yuv *= " << aConfig.mMultiplier << ".0;" << endl;
       }
-      fs << "  vec3 coeff = vec3(0.06275, 0.50196, 0.50196 );" << endl;
-      fs << "  yuv -= coeff;" << endl;
+      fs << "  yuv -= uYuvOffsetVector;" << endl;
       fs << "  color.rgb = uYuvColorMatrix * yuv;" << endl;
       fs << "  color.a = 1.0;" << endl;
     } else if (aConfig.mFeatures & ENABLE_TEXTURE_COMPONENT_ALPHA) {
@@ -854,7 +860,7 @@ bool ShaderProgramOGL::Initialize() {
   NS_ASSERTION(mProgramState == STATE_NEW,
                "Shader program has already been initialised");
 
-  ostringstream vs, fs;
+  std::ostringstream vs, fs;
   for (uint32_t i = 0; i < mProfile.mDefines.Length(); ++i) {
     vs << mProfile.mDefines[i] << endl;
     fs << mProfile.mDefines[i] << endl;
@@ -892,7 +898,7 @@ GLint ShaderProgramOGL::CreateShader(GLenum aShaderType,
    */
   if (!success
 #ifdef DEBUG
-      || (len > 10 && gfxEnv::DebugShaders())
+      || (len > 10 && gfxEnv::MOZ_DEBUG_SHADERS())
 #endif
   ) {
     nsAutoCString log;
@@ -932,9 +938,8 @@ bool ShaderProgramOGL::CreateProgram(const char* aVertexShaderString,
   mGL->fAttachShader(result, vertexShader);
   mGL->fAttachShader(result, fragmentShader);
 
-  for (Pair<nsCString, GLuint>& attribute : mProfile.mAttributes) {
-    mGL->fBindAttribLocation(result, attribute.second(),
-                             attribute.first().get());
+  for (std::pair<nsCString, GLuint>& attribute : mProfile.mAttributes) {
+    mGL->fBindAttribLocation(result, attribute.second, attribute.first.get());
   }
 
   mGL->fLinkProgram(result);
@@ -948,7 +953,7 @@ bool ShaderProgramOGL::CreateProgram(const char* aVertexShaderString,
    */
   if (!success
 #ifdef DEBUG
-      || (len > 10 && gfxEnv::DebugShaders())
+      || (len > 10 && gfxEnv::MOZ_DEBUG_SHADERS())
 #endif
   ) {
     nsAutoCString log;
@@ -990,30 +995,63 @@ GLuint ShaderProgramOGL::GetProgram() {
   return mProgram;
 }
 
-void ShaderProgramOGL::SetBlurRadius(float aRX, float aRY) {
-  float f[] = {aRX, aRY};
-  SetUniform(KnownUniform::BlurRadius, 2, f);
-
-  float gaussianKernel[GAUSSIAN_KERNEL_HALF_WIDTH];
-  float sum = 0.0f;
-  for (int i = 0; i < GAUSSIAN_KERNEL_HALF_WIDTH; i++) {
-    float x = i * GAUSSIAN_KERNEL_STEP;
-    float sigma = 1.0f;
-    gaussianKernel[i] =
-        exp(-x * x / (2 * sigma * sigma)) / sqrt(2 * M_PI * sigma * sigma);
-    sum += gaussianKernel[i] * (i == 0 ? 1 : 2);
-  }
-  for (int i = 0; i < GAUSSIAN_KERNEL_HALF_WIDTH; i++) {
-    gaussianKernel[i] /= sum;
-  }
-  SetArrayUniform(KnownUniform::BlurGaussianKernel, GAUSSIAN_KERNEL_HALF_WIDTH,
-                  gaussianKernel);
-}
-
 void ShaderProgramOGL::SetYUVColorSpace(gfx::YUVColorSpace aYUVColorSpace) {
   const float* yuvToRgb =
       gfxUtils::YuvToRgbMatrix3x3ColumnMajor(aYUVColorSpace);
   SetMatrix3fvUniform(KnownUniform::YuvColorMatrix, yuvToRgb);
+  if (aYUVColorSpace == gfx::YUVColorSpace::Identity) {
+    const float identity[] = {0.0, 0.0, 0.0};
+    SetVec3fvUniform(KnownUniform::YuvOffsetVector, identity);
+  } else {
+    const float offset[] = {0.06275, 0.50196, 0.50196};
+    SetVec3fvUniform(KnownUniform::YuvOffsetVector, offset);
+  }
+}
+
+ShaderProgramOGLsHolder::ShaderProgramOGLsHolder(gl::GLContext* aGL)
+    : mGL(aGL) {}
+
+ShaderProgramOGLsHolder::~ShaderProgramOGLsHolder() { Clear(); }
+
+ShaderProgramOGL* ShaderProgramOGLsHolder::GetShaderProgramFor(
+    const ShaderConfigOGL& aConfig) {
+  auto iter = mPrograms.find(aConfig);
+  if (iter != mPrograms.end()) {
+    return iter->second.get();
+  }
+
+  ProgramProfileOGL profile = ProgramProfileOGL::GetProfileFor(aConfig);
+  auto shader = MakeUnique<ShaderProgramOGL>(mGL, profile);
+  if (!shader->Initialize()) {
+    gfxCriticalError() << "Shader compilation failure, cfg:"
+                       << " features: " << gfx::hexa(aConfig.mFeatures)
+                       << " multiplier: " << aConfig.mMultiplier
+                       << " op: " << aConfig.mCompositionOp;
+    return nullptr;
+  }
+
+  mPrograms.emplace(aConfig, std::move(shader));
+  return mPrograms[aConfig].get();
+}
+
+void ShaderProgramOGLsHolder::Clear() { mPrograms.clear(); }
+
+ShaderProgramOGL* ShaderProgramOGLsHolder::ActivateProgram(
+    const ShaderConfigOGL& aConfig) {
+  ShaderProgramOGL* program = GetShaderProgramFor(aConfig);
+  MOZ_DIAGNOSTIC_ASSERT(program);
+  if (!program) {
+    return nullptr;
+  }
+  if (mCurrentProgram != program) {
+    mGL->fUseProgram(program->GetProgram());
+    mCurrentProgram = program;
+  }
+  return program;
+}
+
+void ShaderProgramOGLsHolder::ResetCurrentProgram() {
+  mCurrentProgram = nullptr;
 }
 
 }  // namespace layers
