@@ -5,66 +5,63 @@
  * found in the LICENSE file.
  */
 
-#include "SkString.h"
+#include "include/core/SkString.h"
 
-#include "SkSafeMath.h"
-#include "SkTo.h"
-#include "SkUtils.h"
+#include "include/private/base/SkDebug.h"
+#include "include/private/base/SkMalloc.h"
+#include "include/private/base/SkTPin.h"
+#include "include/private/base/SkTo.h"
+#include "src/base/SkSafeMath.h"
+#include "src/base/SkUTF.h"
+#include "src/base/SkUtils.h"
 
-#include <cstdarg>
+#include <algorithm>
 #include <cstdio>
+#include <cstring>
 #include <new>
+#include <string_view>
 #include <utility>
 
-
 // number of bytes (on the stack) to receive the printf result
 static const size_t kBufferSize = 1024;
 
-static const char* apply_format_string(const char* format, va_list args, char* stackBuffer,
-                                       size_t stackBufferSize, int* length, SkString* heapBuffer) {
+struct StringBuffer {
+    char*  fText;
+    int    fLength;
+};
+
+template <int SIZE>
+static StringBuffer apply_format_string(const char* format, va_list args, char (&stackBuffer)[SIZE],
+                                        SkString* heapBuffer) SK_PRINTF_LIKE(1, 0);
+
+template <int SIZE>
+static StringBuffer apply_format_string(const char* format, va_list args, char (&stackBuffer)[SIZE],
+                                        SkString* heapBuffer) {
+    // First, attempt to print directly to the stack buffer.
     va_list argsCopy;
     va_copy(argsCopy, args);
-    *length = std::vsnprintf(stackBuffer, stackBufferSize, format, args);
-    if (*length < 0) {
+    int outLength = std::vsnprintf(stackBuffer, SIZE, format, args);
+    if (outLength < 0) {
         SkDebugf("SkString: vsnprintf reported error.");
         va_end(argsCopy);
-        *length = 0;
-        return stackBuffer;
+        return {stackBuffer, 0};
     }
-    if (*length < SkToInt(stackBufferSize)) {
+    if (outLength < SIZE) {
         va_end(argsCopy);
-        return stackBuffer;
+        return {stackBuffer, outLength};
     }
-    heapBuffer->resize(*length);
-    SkDEBUGCODE(int check =)
-            std::vsnprintf(heapBuffer->writable_str(), *length + 1, format, argsCopy);
-    SkASSERT(check == *length);
+
+    // Our text was too long to fit on the stack! However, we now know how much space we need to
+    // format it. Format the string into our heap buffer. `set` automatically reserves an extra
+    // byte at the end of the buffer for a null terminator, so we don't need to add one here.
+    heapBuffer->set(nullptr, outLength);
+    char* heapBufferDest = heapBuffer->data();
+    SkDEBUGCODE(int checkLength =) std::vsnprintf(heapBufferDest, outLength + 1, format, argsCopy);
+    SkASSERT(checkLength == outLength);
     va_end(argsCopy);
-    return heapBuffer->c_str();
+    return {heapBufferDest, outLength};
 }
 
-#define ARGS_TO_BUFFER(format, buffer, size, written, result)                          \
-    SkString overflow;                                                                 \
-    do {                                                                               \
-        va_list args;                                                                  \
-        va_start(args, format);                                                        \
-        result = apply_format_string(format, args, buffer, size, &written, &overflow); \
-        va_end(args);                                                                  \
-    } while (0)
-
-#define V_SKSTRING_PRINTF(output, format)                                                       \
-    do {                                                                                        \
-        char buffer[kBufferSize];                                                               \
-        va_list args;                                                                           \
-        va_start(args, format);                                                                 \
-        int length;                                                                             \
-        auto result = apply_format_string(format, args, buffer, kBufferSize, &length, &output); \
-        SkASSERT(result == output.c_str() || result == buffer);                                 \
-        if (result == buffer) {                                                                 \
-            output.set(buffer, length);                                                         \
-        }                                                                                       \
-    } while (0)
-
 ///////////////////////////////////////////////////////////////////////////////
 
 bool SkStrEndsWith(const char string[], const char suffixStr[]) {
@@ -102,7 +99,7 @@ int SkStrStartsWithOneOf(const char string[], const char prefixes[]) {
 char* SkStrAppendU32(char string[], uint32_t dec) {
     SkDEBUGCODE(char* start = string;)
 
-    char    buffer[SkStrAppendU32_MaxSize];
+    char    buffer[kSkStrAppendU32_MaxSize];
     char*   p = buffer + sizeof(buffer);
 
     do {
@@ -111,11 +108,11 @@ char* SkStrAppendU32(char string[], uint32_t dec) {
     } while (dec != 0);
 
     SkASSERT(p >= buffer);
-    char* stop = buffer + sizeof(buffer);
-    while (p < stop) {
-        *string++ = *p++;
-    }
-    SkASSERT(string - start <= SkStrAppendU32_MaxSize);
+    size_t cp_len = buffer + sizeof(buffer) - p;
+    memcpy(string, p, cp_len);
+    string += cp_len;
+
+    SkASSERT(string - start <= kSkStrAppendU32_MaxSize);
     return string;
 }
 
@@ -131,7 +128,7 @@ char* SkStrAppendS32(char string[], int32_t dec) {
 char* SkStrAppendU64(char string[], uint64_t dec, int minDigits) {
     SkDEBUGCODE(char* start = string;)
 
-    char    buffer[SkStrAppendU64_MaxSize];
+    char    buffer[kSkStrAppendU64_MaxSize];
     char*   p = buffer + sizeof(buffer);
 
     do {
@@ -150,7 +147,7 @@ char* SkStrAppendU64(char string[], uint64_t dec, int minDigits) {
     memcpy(string, p, cp_len);
     string += cp_len;
 
-    SkASSERT(string - start <= SkStrAppendU64_MaxSize);
+    SkASSERT(string - start <= kSkStrAppendU64_MaxSize);
     return string;
 }
 
@@ -163,14 +160,30 @@ char* SkStrAppendS64(char string[], int64_t dec, int minDigits) {
     return SkStrAppendU64(string, udec, minDigits);
 }
 
-char* SkStrAppendFloat(char string[], float value) {
+char* SkStrAppendScalar(char string[], SkScalar value) {
+    // Handle infinity and NaN ourselves to ensure consistent cross-platform results.
+    // (e.g.: `inf` versus `1.#INF00`, `nan` versus `-nan` for high-bit-set NaNs)
+    if (SkScalarIsNaN(value)) {
+        strcpy(string, "nan");
+        return string + 3;
+    }
+    if (!SkScalarIsFinite(value)) {
+        if (value > 0) {
+            strcpy(string, "inf");
+            return string + 3;
+        } else {
+            strcpy(string, "-inf");
+            return string + 4;
+        }
+    }
+
     // since floats have at most 8 significant digits, we limit our %g to that.
     static const char gFormat[] = "%.8g";
     // make it 1 larger for the terminating 0
-    char buffer[SkStrAppendScalar_MaxSize + 1];
+    char buffer[kSkStrAppendScalar_MaxSize + 1];
     int len = snprintf(buffer, sizeof(buffer), gFormat, value);
     memcpy(string, buffer, len);
-    SkASSERT(len <= SkStrAppendScalar_MaxSize);
+    SkASSERT(len <= kSkStrAppendScalar_MaxSize);
     return string + len;
 }
 
@@ -246,17 +259,22 @@ bool SkString::Rec::unique() const {
 }
 
 #ifdef SK_DEBUG
-void SkString::validate() const {
-    // make sure know one has written over our global
+int32_t SkString::Rec::getRefCnt() const {
+    return fRefCnt.load(std::memory_order_relaxed);
+}
+
+const SkString& SkString::validate() const {
+    // make sure no one has written over our global
     SkASSERT(0 == gEmptyRec.fLength);
-    SkASSERT(0 == gEmptyRec.fRefCnt.load(std::memory_order_relaxed));
+    SkASSERT(0 == gEmptyRec.getRefCnt());
     SkASSERT(0 == gEmptyRec.data()[0]);
 
     if (fRec.get() != &gEmptyRec) {
         SkASSERT(fRec->fLength > 0);
-        SkASSERT(fRec->fRefCnt.load(std::memory_order_relaxed) > 0);
+        SkASSERT(fRec->getRefCnt() > 0);
         SkASSERT(0 == fRec->data()[fRec->fLength]);
     }
+    return *this;
 }
 #endif
 
@@ -279,17 +297,18 @@ SkString::SkString(const char text[], size_t len) {
     fRec = Rec::Make(text, len);
 }
 
-SkString::SkString(const SkString& src) {
-    src.validate();
+SkString::SkString(const SkString& src) : fRec(src.validate().fRec) {}
 
-    fRec = src.fRec;
+SkString::SkString(SkString&& src) : fRec(std::move(src.validate().fRec)) {
+    src.fRec.reset(const_cast<Rec*>(&gEmptyRec));
 }
 
-SkString::SkString(SkString&& src) {
-    src.validate();
+SkString::SkString(const std::string& src) {
+    fRec = Rec::Make(src.c_str(), src.size());
+}
 
-    fRec = std::move(src.fRec);
-    src.fRec.reset(const_cast<Rec*>(&gEmptyRec));
+SkString::SkString(std::string_view src) {
+    fRec = Rec::Make(src.data(), src.length());
 }
 
 SkString::~SkString() {
@@ -307,16 +326,12 @@ bool SkString::equals(const char text[]) const {
 bool SkString::equals(const char text[], size_t len) const {
     SkASSERT(len == 0 || text != nullptr);
 
-    return fRec->fLength == len && !memcmp(fRec->data(), text, len);
+    return fRec->fLength == len && !sk_careful_memcmp(fRec->data(), text, len);
 }
 
 SkString& SkString::operator=(const SkString& src) {
     this->validate();
-
-    if (fRec != src.fRec) {
-        SkString    tmp(src);
-        this->swap(tmp);
-    }
+    fRec = src.fRec;  // sk_sp<Rec>::operator=(const sk_sp<Ref>&) checks for self-assignment.
     return *this;
 }
 
@@ -331,11 +346,7 @@ SkString& SkString::operator=(SkString&& src) {
 
 SkString& SkString::operator=(const char text[]) {
     this->validate();
-
-    SkString tmp(text);
-    this->swap(tmp);
-
-    return *this;
+    return *this = SkString(text);
 }
 
 void SkString::reset() {
@@ -343,7 +354,7 @@ void SkString::reset() {
     fRec.reset(const_cast<Rec*>(&gEmptyRec));
 }
 
-char* SkString::writable_str() {
+char* SkString::data() {
     this->validate();
 
     if (fRec->fLength) {
@@ -354,31 +365,40 @@ char* SkString::writable_str() {
     return fRec->data();
 }
 
+void SkString::resize(size_t len) {
+    len = trim_size_t_to_u32(len);
+    if (0 == len) {
+        this->reset();
+    } else if (fRec->unique() && ((len >> 2) <= (fRec->fLength >> 2))) {
+        // Use less of the buffer we have without allocating a smaller one.
+        char* p = this->data();
+        p[len] = '\0';
+        fRec->fLength = SkToU32(len);
+    } else {
+        SkString newString(len);
+        char* dest = newString.data();
+        int copyLen = std::min<uint32_t>(len, this->size());
+        memcpy(dest, this->c_str(), copyLen);
+        dest[copyLen] = '\0';
+        this->swap(newString);
+    }
+}
+
 void SkString::set(const char text[]) {
     this->set(text, text ? strlen(text) : 0);
 }
 
 void SkString::set(const char text[], size_t len) {
     len = trim_size_t_to_u32(len);
-    bool unique = fRec->unique();
     if (0 == len) {
         this->reset();
-    } else if (unique && len <= fRec->fLength) {
-        // should we resize if len <<<< fLength, to save RAM? (e.g. len < (fLength>>1))?
-        // just use less of the buffer without allocating a smaller one
-        char* p = this->writable_str();
+    } else if (fRec->unique() && ((len >> 2) <= (fRec->fLength >> 2))) {
+        // Use less of the buffer we have without allocating a smaller one.
+        char* p = this->data();
         if (text) {
             memcpy(p, text, len);
         }
-        p[len] = 0;
-        fRec->fLength = SkToU32(len);
-    } else if (unique && (fRec->fLength >> 2) == (len >> 2)) {
-        // we have spare room in the current allocation, so don't alloc a larger one
-        char* p = this->writable_str();
-        if (text) {
-            memcpy(p, text, len);
-        }
-        p[len] = 0;
+        p[len] = '\0';
         fRec->fLength = SkToU32(len);
     } else {
         SkString tmp(text, len);
@@ -414,7 +434,7 @@ void SkString::insert(size_t offset, const char text[], size_t len) {
             and we can then eliminate the +1+3 since that doesn't affec the answer
         */
         if (fRec->unique() && (length >> 2) == ((length + len) >> 2)) {
-            char* dst = this->writable_str();
+            char* dst = this->data();
 
             if (offset < length) {
                 memmove(dst + offset + len, dst + offset, length - offset);
@@ -428,7 +448,7 @@ void SkString::insert(size_t offset, const char text[], size_t len) {
                 (we have the original data), and might be faster than alloc/copy/free.
             */
             SkString    tmp(fRec->fLength + len);
-            char*       dst = tmp.writable_str();
+            char*       dst = tmp.data();
 
             if (offset > 0) {
                 memcpy(dst, fRec->data(), offset);
@@ -454,25 +474,25 @@ void SkString::insertUnichar(size_t offset, SkUnichar uni) {
 }
 
 void SkString::insertS32(size_t offset, int32_t dec) {
-    char    buffer[SkStrAppendS32_MaxSize];
+    char    buffer[kSkStrAppendS32_MaxSize];
     char*   stop = SkStrAppendS32(buffer, dec);
     this->insert(offset, buffer, stop - buffer);
 }
 
 void SkString::insertS64(size_t offset, int64_t dec, int minDigits) {
-    char    buffer[SkStrAppendS64_MaxSize];
+    char    buffer[kSkStrAppendS64_MaxSize];
     char*   stop = SkStrAppendS64(buffer, dec, minDigits);
     this->insert(offset, buffer, stop - buffer);
 }
 
 void SkString::insertU32(size_t offset, uint32_t dec) {
-    char    buffer[SkStrAppendU32_MaxSize];
+    char    buffer[kSkStrAppendU32_MaxSize];
     char*   stop = SkStrAppendU32(buffer, dec);
     this->insert(offset, buffer, stop - buffer);
 }
 
 void SkString::insertU64(size_t offset, uint64_t dec, int minDigits) {
-    char    buffer[SkStrAppendU64_MaxSize];
+    char    buffer[kSkStrAppendU64_MaxSize];
     char*   stop = SkStrAppendU64(buffer, dec, minDigits);
     this->insert(offset, buffer, stop - buffer);
 }
@@ -498,50 +518,69 @@ void SkString::insertHex(size_t offset, uint32_t hex, int minDigits) {
 }
 
 void SkString::insertScalar(size_t offset, SkScalar value) {
-    char    buffer[SkStrAppendScalar_MaxSize];
+    char    buffer[kSkStrAppendScalar_MaxSize];
     char*   stop = SkStrAppendScalar(buffer, value);
     this->insert(offset, buffer, stop - buffer);
 }
 
+///////////////////////////////////////////////////////////////////////////////
+
 void SkString::printf(const char format[], ...) {
-    V_SKSTRING_PRINTF((*this), format);
+    va_list args;
+    va_start(args, format);
+    this->printVAList(format, args);
+    va_end(args);
+}
+
+void SkString::printVAList(const char format[], va_list args) {
+    char stackBuffer[kBufferSize];
+    StringBuffer result = apply_format_string(format, args, stackBuffer, this);
+
+    if (result.fText == stackBuffer) {
+        this->set(result.fText, result.fLength);
+    }
 }
 
 void SkString::appendf(const char format[], ...) {
-    char buffer[kBufferSize];
-    int length;
-    const char* result;
-    ARGS_TO_BUFFER(format, buffer, kBufferSize, length, result);
-
-    this->append(result, length);
+    va_list args;
+    va_start(args, format);
+    this->appendVAList(format, args);
+    va_end(args);
 }
 
 void SkString::appendVAList(const char format[], va_list args) {
-    char buffer[kBufferSize];
-    int length = vsnprintf(buffer, kBufferSize, format, args);
-    SkASSERT(length >= 0 && length < SkToInt(kBufferSize));
+    if (this->isEmpty()) {
+        this->printVAList(format, args);
+        return;
+    }
 
-    this->append(buffer, length);
+    SkString overflow;
+    char stackBuffer[kBufferSize];
+    StringBuffer result = apply_format_string(format, args, stackBuffer, &overflow);
+
+    this->append(result.fText, result.fLength);
 }
 
 void SkString::prependf(const char format[], ...) {
-    char buffer[kBufferSize];
-    int length;
-    const char* result;
-    ARGS_TO_BUFFER(format, buffer, kBufferSize, length, result);
-
-    this->prepend(result, length);
+    va_list args;
+    va_start(args, format);
+    this->prependVAList(format, args);
+    va_end(args);
 }
 
 void SkString::prependVAList(const char format[], va_list args) {
-    char buffer[kBufferSize];
-    int length = vsnprintf(buffer, kBufferSize, format, args);
-    SkASSERT(length >= 0 && length < SkToInt(kBufferSize));
+    if (this->isEmpty()) {
+        this->printVAList(format, args);
+        return;
+    }
 
-    this->prepend(buffer, length);
+    SkString overflow;
+    char stackBuffer[kBufferSize];
+    StringBuffer result = apply_format_string(format, args, stackBuffer, &overflow);
+
+    this->prepend(result.fText, result.fLength);
 }
 
-
 ///////////////////////////////////////////////////////////////////////////////
 
 void SkString::remove(size_t offset, size_t length) {
@@ -555,7 +594,7 @@ void SkString::remove(size_t offset, size_t length) {
         SkASSERT(offset <= size - length);
         if (length > 0) {
             SkString    tmp(size - length);
-            char*       dst = tmp.writable_str();
+            char*       dst = tmp.data();
             const char* src = this->c_str();
 
             if (offset) {
@@ -583,37 +622,9 @@ void SkString::swap(SkString& other) {
 
 SkString SkStringPrintf(const char* format, ...) {
     SkString formattedOutput;
-    V_SKSTRING_PRINTF(formattedOutput, format);
+    va_list args;
+    va_start(args, format);
+    formattedOutput.printVAList(format, args);
+    va_end(args);
     return formattedOutput;
 }
-
-void SkStrSplit(const char* str, const char* delimiters, SkStrSplitMode splitMode,
-                SkTArray<SkString>* out) {
-    if (splitMode == kCoalesce_SkStrSplitMode) {
-        // Skip any delimiters.
-        str += strspn(str, delimiters);
-    }
-    if (!*str) {
-        return;
-    }
-
-    while (true) {
-        // Find a token.
-        const size_t len = strcspn(str, delimiters);
-        if (splitMode == kStrict_SkStrSplitMode || len > 0) {
-            out->push_back().set(str, len);
-            str += len;
-        }
-
-        if (!*str) {
-            return;
-        }
-        if (splitMode == kCoalesce_SkStrSplitMode) {
-            // Skip any delimiters.
-            str += strspn(str, delimiters);
-        } else {
-            // Skip one delimiter.
-            str += 1;
-        }
-    }
-}
