@@ -7,13 +7,20 @@
 
 #include "gtest/gtest.h"
 
+#include "gfxTypes.h"
 #include "nsRect.h"
+#include "nsRectAbsolute.h"
 #include "gfxRect.h"
+#include "mozilla/gfx/Rect.h"
+#include "mozilla/gfx/RectAbsolute.h"
 #include "mozilla/WritingModes.h"
 #ifdef XP_WIN
 #  include <windows.h>
 #endif
 
+using mozilla::gfx::IntRect;
+using mozilla::gfx::IntRectAbsolute;
+
 template <class RectType>
 static bool TestConstructors() {
   // Create a rectangle
@@ -429,6 +436,18 @@ static bool TestUnion() {
   return true;
 }
 
+template <class RectType>
+static void TestUnionEmptyRects() {
+  RectType rect1(10, 10, 0, 50);
+  RectType rect2(5, 5, 40, 0);
+  EXPECT_TRUE(rect1.IsEmpty() && rect2.IsEmpty());
+
+  RectType dest = rect1.Union(rect2);
+  EXPECT_TRUE(dest.IsEmpty() && dest.IsEqualEdges(rect2))
+      << "Test the case where both rects are empty, and the result is the "
+         "same value passing into Union()";
+}
+
 static bool TestFiniteGfx() {
   float posInf = std::numeric_limits<float>::infinity();
   float negInf = -std::numeric_limits<float>::infinity();
@@ -517,18 +536,18 @@ static void TestIntersectionLogicalHelper(nscoord x1, nscoord y1, nscoord w1,
   mozilla::LogicalRect r2(mozilla::WritingMode(), rect2.X(), rect2.Y(),
                           rect2.Width(), rect2.Height());
   EXPECT_TRUE(isNonEmpty == r1.IntersectRect(r1, r2));
-  EXPECT_TRUE(rectDebug.IsEqualEdges(
-      nsRect(r1.IStart(WritingMode()), r1.BStart(WritingMode()),
-             r1.ISize(WritingMode()), r1.BSize(WritingMode()))));
+  EXPECT_TRUE(rectDebug.IsEqualEdges(nsRect(
+      r1.IStart(mozilla::WritingMode()), r1.BStart(mozilla::WritingMode()),
+      r1.ISize(mozilla::WritingMode()), r1.BSize(mozilla::WritingMode()))));
 
   mozilla::LogicalRect r3(mozilla::WritingMode(), rect1.X(), rect1.Y(),
                           rect1.Width(), rect1.Height());
   mozilla::LogicalRect r4(mozilla::WritingMode(), rect2.X(), rect2.Y(),
                           rect2.Width(), rect2.Height());
   EXPECT_TRUE(isNonEmpty == r4.IntersectRect(r3, r4));
-  EXPECT_TRUE(rectDebug.IsEqualEdges(
-      nsRect(r4.IStart(WritingMode()), r4.BStart(WritingMode()),
-             r4.ISize(WritingMode()), r4.BSize(WritingMode()))));
+  EXPECT_TRUE(rectDebug.IsEqualEdges(nsRect(
+      r4.IStart(mozilla::WritingMode()), r4.BStart(mozilla::WritingMode()),
+      r4.ISize(mozilla::WritingMode()), r4.BSize(mozilla::WritingMode()))));
 
   mozilla::LogicalRect r5(mozilla::WritingMode(), rect1.X(), rect1.Y(),
                           rect1.Width(), rect1.Height());
@@ -536,9 +555,9 @@ static void TestIntersectionLogicalHelper(nscoord x1, nscoord y1, nscoord w1,
                           rect2.Width(), rect2.Height());
   mozilla::LogicalRect r7(mozilla::WritingMode(), 0, 0, 1, 1);
   EXPECT_TRUE(isNonEmpty == r7.IntersectRect(r5, r6));
-  EXPECT_TRUE(rectDebug.IsEqualEdges(
-      nsRect(r7.IStart(WritingMode()), r7.BStart(WritingMode()),
-             r7.ISize(WritingMode()), r7.BSize(WritingMode()))));
+  EXPECT_TRUE(rectDebug.IsEqualEdges(nsRect(
+      r7.IStart(mozilla::WritingMode()), r7.BStart(mozilla::WritingMode()),
+      r7.ISize(mozilla::WritingMode()), r7.BSize(mozilla::WritingMode()))));
 }
 
 static void TestIntersectionLogical(nscoord x1, nscoord y1, nscoord w1,
@@ -572,6 +591,7 @@ TEST(Gfx, nsRect)
   TestIntersects<nsRect>();
   TestIntersection<nsRect>();
   TestUnion<nsRect>();
+  TestUnionEmptyRects<nsRect>();
   TestBug1135677<nsRect>();
   TestSetWH<nsRect>();
   TestSwap<nsRect>();
@@ -585,6 +605,7 @@ TEST(Gfx, nsIntRect)
   TestIntersects<nsIntRect>();
   TestIntersection<nsIntRect>();
   TestUnion<nsIntRect>();
+  TestUnionEmptyRects<nsIntRect>();
   TestBug1135677<nsIntRect>();
   TestSetWH<nsIntRect>();
   TestSwap<nsIntRect>();
@@ -598,12 +619,19 @@ TEST(Gfx, gfxRect)
   TestIntersects<gfxRect>();
   TestIntersection<gfxRect>();
   TestUnion<gfxRect>();
+  TestUnionEmptyRects<gfxRect>();
   TestBug1135677<gfxRect>();
   TestFiniteGfx();
   TestSetWH<gfxRect>();
   TestSwap<gfxRect>();
 }
 
+TEST(Gfx, nsRectAbsolute)
+{ TestUnionEmptyRects<nsRectAbsolute>(); }
+
+TEST(Gfx, IntRectAbsolute)
+{ TestUnionEmptyRects<IntRectAbsolute>(); }
+
 static void TestMoveInsideAndClamp(IntRect aSrc, IntRect aTarget,
                                    IntRect aExpected) {
   // Test the implementation in BaseRect (x/y/width/height representation)
