@@ -19,7 +19,14 @@ CARGOFLAGS=${CARGOFLAGS:-""}  # default to empty if not set
 WRENCH_BINARY=${WRENCH_BINARY:-""}
 
 pushd wrench
-python script/headless.py reftest
+
+# Test that all shaders compile successfully and pass tests.
+python3 script/headless.py --precache test_init
+python3 script/headless.py --precache --use-unoptimized-shaders test_init
+python3 script/headless.py test_shaders
+
+python3 script/headless.py reftest
+python3 script/headless.py test_invalidation
 if [[ -z "${WRENCH_BINARY}" ]]; then
     cargo build ${CARGOFLAGS} --release
     WRENCH_BINARY="../target/release/wrench"
